#include "mp_simulation.h"


// Check if x[i] is the giving-max element of j'th row
// If so, add i to S
void max_add(double *x, int i, double xp, double *old, set<int> *S, int j){
  if (x[i]+xp>old[j]){
    S[j].clear();
    S[j].insert(i);
    old[j] = x[i]+xp;
  }
  else if (x[i]+xp==old[j]) S[j].insert(i);
}

// calculate current left and right side, and Fi and Gi
void lrfg(int a, int n, int m, map<int,double> **c, map<int,double> &ci, map<int,double> &cv, map<int,double> **d, map<int,double> &di, map<int,double> &dv, double *x, double *L, double *R, set<int> *F, set<int> *G){
  // initialize L, F, G and R
  for (int i=0; i<n+a*n*m+m; i++){
    L[i]  = x[n*m+i];
    F[i].clear();
    F[i].insert(n*m+i);
    R[i] = -numeric_limits<double>::infinity();
    G[i].clear();
  }
  
  // from index 0 to n-1 of L
  for (map<int,double>::iterator itr=ci.begin(); itr!=ci.end(); ++itr){
    max_add(x,n*m+n+a*n*m+m,itr->second,L,F,itr->first);
  }
  // from index 0 to n-1 of R
  for (map<int,double>::iterator its=di.begin(); its!=di.end(); ++its){
    for (int i=0; i<n; i++){
      max_add(x,n*(its->first)+i,its->second,R,G,i);
    }
  }
  
  // from index n to n+a*n*m-1 of L and R
  for (int s=0; s<a; s++){
    // from index n+n*m*s to n+n*m*(s+1)-1 of L and R
    for (int i=0; i<n; i++){
      for (map<int,double>::iterator itr=c[s][i].begin(); itr!=c[s][i].end(); ++itr){
        for (int j=0; j<m; j++){
          max_add(x,n*j+itr->first,itr->second,L,F,n+n*m*s+n*j+i);
        }
      }
    }
    for (int j=0; j<m; j++){
      for (map<int,double>::iterator its=d[s][j].begin(); its!=d[s][j].end(); ++its){
        for (int i=0; i<n; i++){
          max_add(x,n*(its->first)+i,its->second,R,G,n+n*m*s+n*j+i);
        }
      }
    }
  }
  
  // from index n+a*n*m to n+a*n*m+m-1 of L
  for (map<int,double>::iterator itr=cv.begin(); itr!=cv.end(); ++itr){
    for (int j=0; j<m; j++){
      max_add(x,n*j+itr->first,itr->second,L,F,n+a*n*m+j);
    }
  }
  // from index n+a*n*m to n+a*n*m+m-1 of R
  for (map<int,double>::iterator its=dv.begin(); its!=dv.end(); ++its){
    max_add(x,n*m+n+a*n*m+m,its->second,R,G,n+a*n*m+its->first);
  }
}

bool myunion(set<int> &x, set<int> &y){
  bool ret=false;
  for (set<int>::iterator itr=y.begin(); itr!=y.end(); ++itr){
    ret=(x.insert(*itr)).second||ret;
  }
  return(ret);
}

bool xiny(set<int> &x, set<int> &y){
  for (set<int>::iterator itr=x.begin(); itr!=x.end(); ++itr){
    if (y.count(*itr)<1) return(false);
  }
  return (true);
}

double alpha(int index, int a, int n, int m, map<int,double> **c, map<int,double> &ci, map<int,double> &cv, set<int> &P, double *x){
  double alpha;
  if (P.count(n*m+index)<1) alpha=x[n*m+index];
  else alpha=-numeric_limits<double>::infinity();
  int i,j,s;
  
  if (index<n){
    i=index;

    if (P.count(n*m+n+a*n*m+m)<1){
      map<int,double>::iterator itr=ci.find(i);
      if (itr!=ci.end()) alpha=max(alpha,x[n*m+n+a*n*m+m]+itr->second);
    }
  }
  else if (n<=index && index<n+a*n*m){
    index=index-n;
    s=index/(n*m);
    index=index-s*n*m;
    j=index/n;
    i=index-(j*n);
    
    for (map<int,double>::iterator itr=c[s][i].begin(); itr!=c[s][i].end(); ++itr){
      if (P.count(n*j+itr->first)<1) alpha=max(alpha,x[n*j+itr->first]+itr->second);
    }
    
  }
  else if (n+a*n*m<=index){
    j=index-(n+a*n*m);
    
    for (map<int,double>::iterator itr=cv.begin(); itr!=cv.end(); ++itr){
      if (P.count(n*j+itr->first)<1) alpha=max(alpha,x[n*j+itr->first]+itr->second);
    }
  }
  
  return(alpha);
}

double beta(int index, int a, int n, int m, map<int,double> **d, map<int,double> &di, map<int,double> &dv, set<int> &P, double *x){
  double beta;
  beta=-numeric_limits<double>::infinity();
  int i,j,s;

  if (index<n){
    i=index;
    
    for (map<int,double>::iterator itr=di.begin(); itr!=di.end(); ++itr){
      if (P.count(n*(itr->first)+i)<1) beta=max(beta,x[n*(itr->first)+i]+itr->second);
    }
  }
  else if (n<=index && index<n+a*n*m){
    index=index-n;
    s=index/(n*m);
    index=index-s*n*m;
    j=index/n;
    i=index-(j*n);

    for (map<int,double>::iterator itr=d[s][j].begin(); itr!=d[s][j].end(); ++itr){
      if (P.count(n*(itr->first)+i)<1) beta=max(beta,x[n*(itr->first)+i]+itr->second);
    }
  }
  else if (n+a*n*m<=index){
    j=index-(n+a*n*m);
    
    if (P.count(n*m+n+a*n*m+m)<1){
      map<int,double>::iterator itr=dv.find(j);
      if (itr!=dv.end()) beta=max(beta,x[n*m+n+a*n*m+m]+itr->second);
    }
  }
  
  return(beta);
}

// -inf+inf=-inf
double tminus(double x, double y){
  if (x==-numeric_limits<double>::infinity()) return(x);
  else return (x-y);
}

// calculate -tau
double get_tau(int a, int n, int m, map<int,double> **c, map<int,double> &ci, map<int,double> &cv, map<int,double> **d, map<int,double> &di, map<int,double> &dv, double *x, double *L, double *R, set<int> *F, set<int> *G, set<int> &P){
  double tau=-numeric_limits<double>::infinity();
  
  for (int i=0; i<n+a*n*m+m; i++){
    if (xiny(F[i],P)) tau=max(tau,tminus(alpha(i,a,n,m,c,ci,cv,P,x),L[i]));
    else if (L[i]<R[i]) tau=max(tau,tminus(alpha(i,a,n,m,c,ci,cv,P,x),R[i]));
    if (xiny(G[i],P)) tau=max(tau,tminus(beta(i,a,n,m,d,di,dv,P,x),R[i]));
    else if (R[i]<L[i]) tau=max(tau,tminus(beta(i,a,n,m,d,di,dv,P,x),L[i]));
  }
  
  return (tau);
}



// collect vairables to be decreased
void algorithmI(int a, int n, int m, map<int,double> **c, map<int,double> &ci, map<int,double> &cv, map<int,double> **d, map<int,double> &di, map<int,double> &dv, double *x, set<int> &P,double *L, double *R, set<int> *F, set<int> *G){
  set<int> Q;
  bool itr_end=true;
  // initialize P (P(x)=A(x))
  P.clear();
  for(int i=0; i<n+a*n*m+m; i++){
    Q=F[i];
    if (L[i]>R[i]) myunion(P,F[i]);
    else if (L[i]<R[i]) myunion(P,G[i]);
  }
  while (itr_end){
    // calculate E1 and E2
    itr_end = false;
    for (int i=0; i<n+a*n*m+m; i++){
      if (L[i]==R[i]){
        // if Fi in P then add variables in Gi to P
        if (!F[i].empty()&&xiny(F[i],P)) itr_end=itr_end||myunion(P,G[i]);
        else if (!G[i].empty()&&xiny(G[i],P)) itr_end=itr_end||myunion(P,F[i]);
      }
    }
  }
}

int algorithmII(int a, int n, int m, map<int,double> **c, map<int,double> &ci, map<int,double> &cv, map<int,double> **d, map<int,double> &di, map<int,double> &dv, double *x,set<int> *F, set<int> *G, int iter=-1){
  bool ret;
  double L[n+a*n*m+m],R[n+a*n*m+m];
  double mtau;
  set<int> P,Q;
  //int le1=0,le2=INT_MAX;
  
  // initialize x
  for (int i=0; i<n*m+n+a*m*n+m+1; i++) x[i] = 0;
  
  while(--iter!=0){
    // for (int i=0; i<n*m+n+a*n*m+m+1; i++) cerr << x[i] << " "; cerr <<endl;
    // calculate H, F, G
    lrfg(a, n, m, c, ci, cv, d, di, dv, x, L, R, F, G);
  
    // if H=emptyset, then exit
    ret = true;
    //le1=0;
    for (int i=0; i<n+a*n*m+m; i++){
      if (L[i]!=R[i]){
        ret=false;
        //cerr << i << ":"<< L[i] << ","<< R[i] << " ";

        //le1++;//
        break;
      }
    }//cerr << endl;
    //if (le1!=le2) cerr << le1 << endl; le2=le1;
    if (ret) return(SIMULATION_FOUND);

    // call algorithm I
    algorithmI(a, n, m, c, ci, cv, d, di, dv, x, P, L, R, F, G);


    // if P covers all variables, then exit
    if (P.size()>=n*m+n+a*n*m+m+1) return(SIMULATION_NOTFOUND);
  
    // calculate -tau
    mtau=get_tau(a,n,m,c,ci,cv,d,di,dv,x,L,R,F,G,P);
    //cerr << mtau << endl;

    // add mtau from P
    for (set<int>::iterator itr=P.begin(); itr!=P.end(); ++itr){
      x[*itr]+=mtau;
    }
    if (x[n*m+n+a*n*m+m]==-numeric_limits<double>::infinity()) return(SIMULATION_NOTFOUND);
  }
  
  return (ITERATION_LIMIT);
}



// be careful that c is transposed
// solve tropical linear inequalities
int solve(int a, int n, int m, map<int,double> **c, map<int,double> &ci, map<int,double> &cv, map<int,double> **d, map<int,double> &di, map<int,double> &dv, double *ret, int flags, int iter=-1){
  map<int,double>::iterator itr,its;
  
  // allocate memory
  set<int> *F, *G;
  F=new set<int> [n+a*n*m+m];
  G=new set<int> [n+a*n*m+m];
  double x[n*m+n+a*n*m+m+1];
  
  int retval = algorithmII(a,n,m,c,ci,cv,d,di,dv,x,F,G,iter);
  
  if (flags&OUTPUT) for (int i=0; i<n*m; i++) ret[i]=x[i];
 
  delete [] F;
  delete [] G;
  return (retval);
}